#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <windows.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "raster.h"

//----------------------------------------------------------------------------

static int      iwidth      = 320;
static int      iheigh      = 240;
static int      iantialias  = 1;
static int      tantialias  = 1;
static float    framerate   = 24.0f;
static int      numframes   = 1;

typedef struct
{
    float  b;
    float  g;
    float  r;
}COLOR;

//----------------------------------------------------------------------------
static int mfsize( int h )
{
    unsigned long    hi;
    return( GetFileSize((HANDLE)h,&hi) );    
}

static void * mfopen( char *name, int read )
{
    HANDLE h;

    if( read )
        h = CreateFile( name, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
    else
        h = CreateFile( name, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
    if( h==INVALID_HANDLE_VALUE )
        return( 0 );
    return( (void*)h );
}

static void mfwrite( void *buf, int n, int m, void * h )
{
    unsigned long wr;

    WriteFile( (HANDLE)h, buf, n*m, &wr, 0 );
}

static void mfclose( void * h )
{
    CloseHandle( (HANDLE)h );
}

static int BMP_Salva24( char *name, unsigned char *buffer, int xres, int yres )
{
    unsigned char bmpheader[54] = {
    // Header = 14 bytes
    0x42, 0x4D,              //  0: signature = BM
    0x00, 0x00, 0x00, 0x00,  //  2: [size]
    0x00, 0x00,              //  6: reserved
    0x00, 0x00,              //  8: reserved
    0x36, 0x00, 0x00, 0x00,  // 10: offset to pixels = 54 = sizeof(Header) + sizeof(Info)

    // Info = 40 bytes
    0x28, 0x00, 0x00, 0x00,  // 14: header size = 40
    0x00, 0x00, 0x00, 0x00,  // 18: [xres]
    0x00, 0x00, 0x00, 0x00,  // 22: [yres]
    0x01, 0x00,              // 26: bit planes = 1
    0x18, 0x00,              // 28: bits per pixel = 24
    0x00, 0x00, 0x00, 0x00,  // 30: compression:  0=none, 1=8bit RLE, 2=4bit RLE, 3=RGB with mask
    0x00, 0x00, 0x00, 0x00,  // 34: [image size]
    0xC3, 0x0E, 0x00, 0x00,  // 38: Pixels per meter X (3779)
    0xC3, 0x0E, 0x00, 0x00,  // 42: Pixels per meter Y
    0x00, 0x00, 0x00, 0x00,  // 46: number of colors (pallette)
    0x00, 0x00, 0x00, 0x00   // 50: important colors (pallette)
    };

    void * fo = mfopen( name, 0 );
    if( !fo )
        return( 0 );

    *((int*)(bmpheader+ 2)) = xres*yres*3 + 54;
    *((int*)(bmpheader+18)) = xres;
    *((int*)(bmpheader+22)) = yres;
    *((int*)(bmpheader+34)) = xres*yres*3;
    
    mfwrite(  bmpheader, 1, 54, fo );

    #if 0
    for( int j=0; j<yres; j++ )
    {
        unsigned char *ptr = buffer + 3*(xres*(yres-1-j) );
        mfwrite( ptr, 1, 3*xres, fo);
    }
    #else
    mfwrite( buffer, 1, 3*xres*yres, fo);
    #endif

    mfclose( fo );

    return( 1 );
}


//----------------------------------------------------------------------------

static void clear( COLOR *buffer )
{
    memset( buffer, 0, iwidth*iheigh*iantialias*iantialias*sizeof(COLOR) );
}

static void acum( COLOR *bufferd, COLOR *buffero )
{
    int    l = iwidth*iheigh*3*iantialias*iantialias;
    float  *d = (float*)bufferd;
    float  *s = (float*)buffero;

    while( l-- )
        *d++ += *s++;
}

static void scale( COLOR *buffer )
{
    long    l  = iwidth*iheigh*3*iantialias*iantialias;
    float  *b = (float*)buffer;
    float  s  = 1.0f/(float)tantialias;

    if( tantialias==1 ) return;

    while( l-- )
        *b++ *= s;
}


static void filter( COLOR *dbuffer, COLOR *obuffer )
{
    int   i, j, n, m;
    float  r, g, b, s;

    s = 1.0f/(float)(iantialias*iantialias);

    for( j=0; j<iheigh; j++ )
    for( i=0; i<iwidth; i++ )
        {
        r = g = b = 0.0f;
        for( m=0; m<iantialias; m++ )
        for( n=0; n<iantialias; n++ )
            {
            r += obuffer[ (iantialias*j+m)*iwidth*iantialias + (iantialias*i+n) ].r;
            g += obuffer[ (iantialias*j+m)*iwidth*iantialias + (iantialias*i+n) ].g;
            b += obuffer[ (iantialias*j+m)*iwidth*iantialias + (iantialias*i+n) ].b;
            }

        dbuffer->r = r*s;
        dbuffer->g = g*s;
        dbuffer->b = b*s;
        dbuffer++;
        }
}

static void quantify( unsigned char *rbuffer, COLOR *cbuffer )
{
    long           l  = iwidth*iheigh;
    float          *s = (float*)cbuffer;
    unsigned char  *d = rbuffer;
    int           c;

    while( l-- )
        {
        c = (int)(256.0f* (*s++) ); if( c>255 ) c=255; if( c<0 ) c=0; d[2] = c;
        c = (int)(256.0f* (*s++) ); if( c>255 ) c=255; if( c<0 ) c=0; d[1] = c;
        c = (int)(256.0f* (*s++) ); if( c>255 ) c=255; if( c<0 ) c=0; d[0] = c;
        d += 3;
        }
}


//----------------------------------------------------------------------------

static char str[64];

static void hace_nombre( char *nombre, int i )
{
    nombre[15] = 'p';
    nombre[14] = 'm';
    nombre[13] = 'b';
    nombre[12] = '.';
    nombre[11] = '0' + (i/1)%10;
    nombre[10] = '0' + (i/10)%10;
    nombre[ 9] = '0' + (i/100)%10;
    nombre[ 8] = '0' + (i/1000)%10;
    nombre[ 7] = '0' + (i/10000)%10;
    nombre[ 6] = '0' + (i/100000)%10;
    nombre[ 5] = '0' + (i/1000000)%10;
    nombre[ 4] = '0' + (i/10000000)%10;

    nombre[ 3] = '/';
    nombre[ 2] = 't';
    nombre[ 1] = 'u';
    nombre[ 0] = 'o';

}


static int ya_existe( char *name )
{
    void * h = mfopen( name, 1 );
    if( h )
    {
        mfclose( h );
        return( 1 );
    }

    return( 0 );
}


int RASTER_Config( int width, int heigh, int ian,
                   int tan, float fr, int nf )
{
    iwidth      = width;
    iheigh      = heigh;
    iantialias  = ian;
    tantialias  = tan;
    framerate   = fr;
    numframes   = nf;
    
    if( numframes<0 )
        return( 0 );

    if( iantialias<1 || iantialias>32 )
        return( 0 );

    if( tantialias<1 || tantialias>32 )
        return( 0 );

    if( iwidth<40 || iwidth>8000 )
        return( 0 );

    if( iheigh<30 || iheigh>6000 )
        return( 0 );

    return( 1 );
}


int RASTER_Do( DOFRAME doframe )
{
    unsigned char  *rbuffer;       // raster buffer (24bpp)
    COLOR   *cbuffer_f;     // floating( iantialas )
    COLOR   *cbuffer_a;     // accumulation (tantialas)
    COLOR   *cbuffer_r;     // render

    float   t;
    long    i, j;
    long    cuantos=0;
	int		done = 0;

    if( !doframe )
        return( 0 );

    rbuffer = (unsigned char*)malloc( iwidth*iheigh*3 );
    if( !rbuffer )
		{
        return( 0 );
		}

    cbuffer_f = (COLOR*)malloc( iwidth*iheigh*sizeof(COLOR) );
    if( !cbuffer_f )
        {
        return( 0 );
		}

    cbuffer_a = (COLOR*)malloc( iwidth*iheigh*iantialias*iantialias*sizeof(COLOR) );
    if( !cbuffer_a )
        {
        return( 0 );
		}


    cbuffer_r = (COLOR*)malloc( iwidth*iheigh*iantialias*iantialias*sizeof(COLOR) );
    if( !cbuffer_r )
        {
        return( 0 );
		}


    //-----------------------

    for( i=0; (i<numframes) && !done; i++ )
        {
        hace_nombre( str, i );

        if( !ya_existe( str ) || numframes==1 )
            {
            clear( cbuffer_a );

            for( j=0; j<tantialias; j++ )
                {
                t = (float)i + (float)j/tantialias;
                t = t/framerate;

                done = doframe( (float*)cbuffer_r, iwidth*iantialias, iheigh*iantialias, t );

                acum( cbuffer_a, cbuffer_r );
                }

            scale( cbuffer_a );

            filter( cbuffer_f, cbuffer_a );

            //desaturate( cbuffer_f );
      
            quantify( rbuffer, cbuffer_f );

            if( !BMP_Salva24( str, rbuffer, iwidth, iheigh ) )
                {
                free( rbuffer );
                free( cbuffer_f );
                free( cbuffer_a );
                free( cbuffer_r );
                return( 0 );
                }

            cuantos++;
            }
        }

    free( rbuffer );
    free( cbuffer_f );
    free( cbuffer_a );
    free( cbuffer_r );

    return( 1 );
}
